"""
DC3-MWCP server--simple REST API using the Flask framework.
"""
from __future__ import absolute_import, print_function

import base64
import hashlib
import io
import json
import logging
import os
import zipfile
from copy import copy

import flask as f
import mwcp
import mwcp.parsers
import pygments
import pygments.formatter
import pygments.lexer
from mwcp.utils import logutil
from pygments.formatters.html import HtmlFormatter
from pygments.lexers.data import JsonLexer
from pygments.lexers.special import TextLexer
from werkzeug.utils import secure_filename

bp = f.Blueprint("mwcp", __name__)


def init_app(app):
    """Initialize the Flask application and MWCP"""
    # Initialize MWCP
    mwcp.config.load(os.getenv("MWCP_CONFIG", None))
    mwcp.register_entry_points()


@bp.route("/run_parser/<parser>", methods=["POST"], strict_slashes=False)
@bp.route("/run_parser", methods=["POST"], strict_slashes=False)
def run_parser(parser=None):
    """
    Execute a parser and return the results.

    The file must be uploaded in the field "data".

    The parser must be specified in the `parser` field, parameter, or in the resource name.

    The field `output` may be set to `zip` to download a ZIP
    file of the results and extracted components. By default the
    output is JSON.

    The field  or parameter `highlight` may also be set to return
    a formatted HTML page with the results.

    All fields (except `data`) may be set as URL parameters as well.

    :param str parser: The name of the parser to run
    """
    return _build_parser_response(parser)


@bp.route("/run_parsers/<path:parsers>", methods=["POST"])
def run_parsers(parsers):
    """
    Execute multiple parsers on the same input file.

    NOTE: The way this function works may change in a future version

    :param str parsers: List of parsers to run
    """
    output = {}

    dep_warning = (
        "Running multiple parsers in a single request will be changed future version."
    )
    f.current_app.logger.warning(dep_warning)
    output.setdefault("errors", []).append(dep_warning)

    datafile = f.request.files.get("data")
    if datafile:
        data = datafile.read()
        f.current_app.logger.info(
            "run_parsers %s %s %s",
            parsers,
            datafile.filename,
            hashlib.md5(data).hexdigest(),
        )
        for parser in parsers.split("/"):
            if parser:
                output[parser] = _run_parser(parser, data=data)
    else:
        output.setdefault("errors", []).append("No input file provided")
        f.current_app.logger.error("run_parsers %s no input file", parsers)
    return f.jsonify(output)


@bp.route("/parsers")
def parsers_list():
    """
    List of configured parsers with names, sources, authors, and descriptions.

    Normally an HTML table, but if `application/json` is the best mimetype set
    in the `Accept` header, the response will be in JSON.
    """
    name_filter = f.request.args.get("name", type=str)
    source_filter = f.request.args.get("source", type=str)

    headers = ("Name", "Source", "Author", "Description")
    parsers_info = mwcp.get_parser_descriptions(name=name_filter, source=source_filter)

    if f.request.accept_mimetypes.best == "application/json":
        return f.jsonify(
            {"parsers": [parser_info._asdict() for parser_info in parsers_info]}
        )

    f.g.title = "Parsers"
    return f.render_template("parsers.html", headers=headers, parsers=parsers_info)


@bp.route("/upload")
def upload():
    """Upload page"""
    f.g.title = "Upload"
    parsers_info = mwcp.get_parser_descriptions()
    return f.render_template("upload.html", parsers=parsers_info)


@bp.route("/descriptions")
def descriptions():
    """
    List descriptions of parser modules.
    This is for backwards compatibility purposes.
    Always a JSON response.
    """
    return f.jsonify(
        [
            (parser_info.name, parser_info.author, parser_info.description)
            for parser_info in mwcp.get_parser_descriptions()
        ]
    )


@bp.route("/logs")
def logs():
    """
    Endpoint for all logs from the current session.

    Always a JSON response.

    This can be disabled with the ``DISABLE_LOGS_ENDPOINT`` key
    in the app config.
    """
    if f.current_app.config.get("DISABLE_LOGS_ENDPOINT"):
        return (
            f.jsonify({"errors": ["Logs endpoint has been disabled by configuration"]}),
            403,
        )

    handler = _get_existing_handler()
    if not handler:
        return (
            f.jsonify({"errors": ["No 'mwcp_server' handler defined on root logger."]}),
            500,
        )
    return f.jsonify({"logs": handler.messages})


@bp.route("/")
def default():
    """
    Homepage endpoint.
    """
    return f.render_template("base.html")


class RequestFilter(object):
    """
    Filter to lock a handler to a specific request.

    This is required for multi-threading the server.
    """

    def __init__(self, request=None):
        # The `request` is usually a proxy to the real request
        if hasattr(request, "_get_current_object"):
            request = request._get_current_object()

        self._request = request

    def filter(self, record):
        # Ensure the proxied request is our locked request
        # And the record is created in a request context to begin with
        if not f.has_request_context():
            return False
        return f.request == self._request


def _get_existing_handler(handler_name="mwcp_server"):
    """
    Retrieve an existing ListHandler by name from the root logger.
    """
    for handler in logging.root.handlers:
        if handler.name == handler_name and isinstance(handler, logutil.ListHandler):
            return handler


def _get_log_handler(handler_name="mwcp_server"):
    """
    Get the handler for the parser running.

    Attempts to get 'mwcp_server' handler from the root logger, and create
    a clean copy, keeping any formatters and level settings.

    If the handler does not exist, create a default handler.
    """
    handler = _get_existing_handler(handler_name)
    if handler:
        if isinstance(handler, logutil.ListHandler):
            new_handler = copy(handler)
            new_handler.clear()
            return new_handler
        f.current_app.logger.warning(
            "Root handler '{}' is not a ListHandler.".format(handler_name)
        )

    f.current_app.logger.info(
        "No 'mwcp_server' handler defined on root logger. Using default."
    )
    list_handler = logutil.ListHandler()
    list_handler.setFormatter(
        logging.Formatter("[%(level_char)s] (%(name)s): %(message)s")
    )
    list_handler.addFilter(logutil.LevelCharFilter())

    return list_handler


def _highlight(data, is_json=True):
    """
    Render an HTML page with a highlighted JSON string or plain text.

    :param data: Data to highlight, should be a string or JSON-able object
    :param is_json: If the data is a JSON string or can be converted into such
    :return: Response object with rendered template with highlighted data
    """
    if is_json and not isinstance(data, (str, bytes)):
        data = json.dumps(data, indent=2)

    # Pygments highlighting
    lexer = JsonLexer() if is_json else TextLexer()
    formatter = HtmlFormatter()
    highlight = pygments.highlight(data, lexer, formatter)

    return f.render_template(
        "results.html", highlight=highlight, extra_css=formatter.get_style_defs()
    )


def _build_zip(parser_results):
    """
    Build a ZIP file containing the results and artifacts of a parser run.

    Expects the **full** parser results, including ``output_text`` and ``outputfile`` keys.

    The folder structure looks like this:

    .. code_block::

        mwcp_server.zip
        |
        |-results.json
        |-results.txt (this is ``output_text``)
        |
        |---files
            |
            |- ExtractedComponent1.exe
            |- ExtractedComponent2.dll


    :param parser_results:
    :return: A BytesIO buffer containing a ZIP file
    :rtype: io.BytesIO
    """
    zip_buf = io.BytesIO()

    encoded_files = parser_results.pop("outputfile", [])
    output_text = parser_results.pop("output_text", "")

    zf = zipfile.ZipFile(
        zip_buf, mode="w", compression=zipfile.ZIP_DEFLATED, allowZip64=True
    )
    with zf:
        for file_obj in encoded_files:
            filename = file_obj[0]
            base64_data = file_obj[3]
            file_data = base64.b64decode(base64_data)
            zf.writestr(os.path.join("files", filename), file_data)

        zf.writestr("results.json", json.dumps(parser_results, indent=2))

        if not isinstance(output_text, bytes):
            output_text = output_text.encode("ascii", "backslashreplace")
        zf.writestr("results.txt", output_text)

    zip_buf.seek(0)
    return zip_buf


def _build_parser_response(parser=None, **kwargs):
    """
    Build the response object for a parser request.
    This function handles the form fields and/or URL parameters and
    returns an appropriate response object. This can be overridden
    (e.g. by specific endpoints) as a parameter.

    :param str parser: The name of the parser to run. Pulled from `parser`
        URL parameter or form field if not specified.
    :return: Flask response object
    """
    output = kwargs.get("output", "") or f.request.values.get("output", "json")
    output = output.lower()
    if output not in ("json", "text", "zip"):
        f.current_app.logger.warning(
            "Unknown output type received: '{}'".format(output)
        )
        output = "json"
    highlight = kwargs.get("highlight") or f.request.values.get("highlight")

    if not highlight:
        json_response = f.jsonify
    else:
        json_response = _highlight

    parser_results, response_code = _run_parser_request(parser)

    if response_code != 200:
        return json_response(parser_results), response_code

    # A ZIP returns both JSON and plain text, and has no highlighting
    if output == "zip":
        filename = secure_filename(f.request.files.get("data").filename)
        zip_buf = _build_zip(parser_results)
        return f.send_file(
            zip_buf, "application/zip", True, "{}_mwcp_output.zip".format(filename)
        )

    if highlight:
        parser_results.pop("outputfile", [])
        output_text = parser_results.pop("output_text", "")
        if output == "text":
            return _highlight(output_text, False)

    return json_response(parser_results)


def _run_parser_request(parser=None, upload_name="data", output_text=True):
    """
    Run a parser based on the data in the current request.

    This function handles getting the file from the form field, as well as
    the parser from either a form field or url parameter if not explicitly set.

    The results from the parser run (a ``dict``) is returned as well as an
    appropriate HTTP status code. Specifically, a 2XX if the parser ran
    successfully, a 4XX if there is a problem with the request (e.g no
    file) or a 5XX if there was a problem with running the parser.

    :param str parser: The name of the parser to run. Pulled from `parser`
        URL parameter or form field if not specified.
    :param str upload_name: The name of the field of the uploaded sample
    :param bool output_text: If the `output_text` key should be included in the output
    :return: The results from the parser run and/or errors and an appropriate status code
    :rtype: (dict, int)
    """
    errors = []

    parser = parser or f.request.values.get("parser")
    if not parser:
        errors.append("No parser specified")

    uploaded_file = f.request.files.get(upload_name)
    if not uploaded_file:
        f.current_app.logger.error(
            "Error running parser '{}' no input file".format(parser)
        )
        errors.append("No input file provided")

    # Client errors
    if errors:
        return {"errors": errors}, 400

    data = uploaded_file.read()
    f.current_app.logger.info(
        "Request for parser '%s' on '%s' %s",
        parser,
        secure_filename(uploaded_file.filename),
        hashlib.md5(data).hexdigest(),
    )
    parser_results = _run_parser(parser, data=data, append_output_text=output_text)

    return parser_results, 200


def _run_parser(name, data=b"", append_output_text=True):
    """
    Run an MWCP parser on given data.

    Logs to a list handler that is locked to the current request.

    :param str name: Name of the parser to run
    :param bytes data: Data to run parser on
    :param bool append_output_text: If the text that would otherwise be printed is
        added to the output data
    :return: Output from the reporter
    :rtype: dict
    """
    output = {}
    mwcp_logger = logging.getLogger("mwcp")
    list_handler = _get_log_handler()
    try:
        # Record only records created in the context of this request
        list_handler.addFilter(RequestFilter(request=f.request))
        mwcp_logger.addHandler(list_handler)

        reporter = mwcp.Reporter(base64_output_files=True)
        reporter.run_parser(name, data=data)
        output = reporter.metadata

        output["debug"] = [msg for msg in list_handler.messages]

        if append_output_text:
            output["output_text"] = reporter.get_output_text()
    except Exception as e:
        output = {"errors": [str(e)]}
        if f.has_app_context():
            f.current_app.logger.exception(
                "Error running parser '%s': %s", name, str(e)
            )
    finally:
        mwcp_logger.removeHandler(list_handler)
        return output



#!/usr/bin/env python
"""
A framework for malware configuration parsers.
"""

from setuptools import setup, find_packages

setup(
    name="mwcp",
    author="DC3",
    author_email="dcci@dc3.mil",
    keywords="malware",
    url="http://github.com/Defense-Cyber-Crime-Center/DC3-MWCP/",
    packages=find_packages(),
    include_package_data=True,
    license='MIT',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
    ],
    entry_points={
        'console_scripts': [
            'mwcp = mwcp.cli:main',
            'poshdeob = mwcp.utils.poshdeob:main'
        ],
        'mwcp.parsers': [
            'mwcp = mwcp.parsers',
        ]
    },
    install_requires=[
        'appdirs',
        'click',
        'construct==2.9.45',  # pin because parsers are very dependent on this.
        'future',
        'jinja2',  # For construct.html_hex()
        'pefile>=2019.4.18',
        'pyelftools',
        'pyparsing==2.3.0',  # 2.4.0 seems to break poshdeob
        'pyyaml',
        'requests',
        'ruamel.yaml',
        'six',
        'tabulate',

        # For the server and API
        'flask~=1.1.0',
        'pygments~=2.2.0',
    ],
    extras_require={
        'kordesii': ['kordesii>=2.0.0'],
        'testing': [
            'pytest',
            'pytest-console-scripts',
        ],
    }
)





# IMAGE_SECTION_HEADER.Characteristics
IMAGE_SCN_TYPE_NO_PAD = 'IMAGE_SCN_TYPE_NO_PAD'
IMAGE_SCN_CNT_CODE = 'IMAGE_SCN_CNT_CODE'
IMAGE_SCN_CNT_INITIALIZED_DATA = 'IMAGE_SCN_CNT_INITIALIZED_DATA'
IMAGE_SCN_CNT_UNINITIALIZED_DATA = 'IMAGE_SCN_CNT_UNINITIALIZED_DATA'
IMAGE_SCN_LNK_OTHER = 'IMAGE_SCN_LNK_OTHER'
IMAGE_SCN_LNK_INFO = 'IMAGE_SCN_LNK_INFO'
IMAGE_SCN_LNK_REMOVE = 'IMAGE_SCN_LNK_REMOVE'
IMAGE_SCN_LNK_COMDAT = 'IMAGE_SCN_LNK_COMDAT'
IMAGE_SCN_NO_DEFER_SPEC_EXC = 'IMAGE_SCN_NO_DEFER_SPEC_EXC'
IMAGE_SCN_GPREL = 'IMAGE_SCN_GPREL'
IMAGE_SCN_MEM_PURGEABLE = 'IMAGE_SCN_MEM_PURGEABLE'
IMAGE_SCN_MEM_LOCKED = 'IMAGE_SCN_MEM_LOCKED'
IMAGE_SCN_MEM_PRELOAD = 'IMAGE_SCN_MEM_PRELOAD'
IMAGE_SCN_ALIGN_1BYTES = 'IMAGE_SCN_ALIGN_1BYTES'
IMAGE_SCN_ALIGN_2BYTES = 'IMAGE_SCN_ALIGN_2BYTES'
IMAGE_SCN_ALIGN_4BYTES = 'IMAGE_SCN_ALIGN_4BYTES'
IMAGE_SCN_ALIGN_8BYTES = 'IMAGE_SCN_ALIGN_8BYTES'
IMAGE_SCN_ALIGN_16BYTES = 'IMAGE_SCN_ALIGN_16BYTES'
IMAGE_SCN_ALIGN_32BYTES = 'IMAGE_SCN_ALIGN_32BYTES'
IMAGE_SCN_ALIGN_64BYTES = 'IMAGE_SCN_ALIGN_64BYTES'
IMAGE_SCN_ALIGN_128BYTES = 'IMAGE_SCN_ALIGN_128BYTES'
IMAGE_SCN_ALIGN_256BYTES = 'IMAGE_SCN_ALIGN_256BYTES'
IMAGE_SCN_ALIGN_512BYTES = 'IMAGE_SCN_ALIGN_512BYTES'
IMAGE_SCN_ALIGN_1024BYTES = 'IMAGE_SCN_ALIGN_1024BYTES'
IMAGE_SCN_ALIGN_2048BYTES = 'IMAGE_SCN_ALIGN_2048BYTES'
IMAGE_SCN_ALIGN_4096BYTES = 'IMAGE_SCN_ALIGN_4096BYTES'
IMAGE_SCN_ALIGN_8192BYTES = 'IMAGE_SCN_ALIGN_8192BYTES'
IMAGE_SCN_LNK_NRELOC_OVFL = 'IMAGE_SCN_LNK_NRELOC_OVFL'
IMAGE_SCN_MEM_DISCARDABLE = 'IMAGE_SCN_MEM_DISCARDABLE'
IMAGE_SCN_MEM_NOT_CACHED = 'IMAGE_SCN_MEM_NOT_CACHED'
IMAGE_SCN_MEM_NOT_PAGED = 'IMAGE_SCN_MEM_NOT_PAGED'
IMAGE_SCN_MEM_SHARED = 'IMAGE_SCN_MEM_SHARED'
IMAGE_SCN_MEM_EXECUTE = 'IMAGE_SCN_MEM_EXECUTE'
IMAGE_SCN_MEM_READ = 'IMAGE_SCN_MEM_READ'
IMAGE_SCN_MEM_WRITE = 'IMAGE_SCN_MEM_WRITE'

# IMAGE_OPTIONAL_HEADER.Magic
IMAGE_NT_OPTIONAL_HDR32_MAGIC = 0x10b
IMAGE_NT_OPTIONAL_HDR64_MAGIC = 0x20b
IMAGE_ROM_OPTIONAL_HDR_MAGIC = 0x107

# IMAGE_OPTIONAL_HEADER.Subsystem
IMAGE_SUBSYSTEM_UNKNOWN = 0
IMAGE_SUBSYSTEM_NATIVE = 1
IMAGE_SUBSYSTEM_WINDOWS_GUI = 2
IMAGE_SUBSYSTEM_WINDOWS_CUI = 3
IMAGE_SUBSYSTEM_OS2_CUI = 5
IMAGE_SUBSYSTEM_POSIX_CUI = 7
IMAGE_SUBSYSTEM_WINDOWS_CE_GUI = 9
IMAGE_SUBSYSTEM_EFI_APPLICATION = 10
IMAGE_SUBSYSTEM_EFI_BOOT_SERVICE_DRIVER = 11
IMAGE_SUBSYSTEM_EFI_RUNTIME_DRIVER = 12
IMAGE_SUBSYSTEM_EFI_ROM = 13
IMAGE_SUBSYSTEM_XBOX = 14
IMAGE_SUBSYSTEM_WINDOWS_BOOT_APPLICATION = 16


# Make default DataDirectory and standard indexes available for convenience.
# WARNING: Make sure you make a copy of DEFAULT_DATA_DIRECTORIES!!
DATA_DIR_INDEX_EXPORTS = 0
DATA_DIR_INDEX_IMPORTS = 1
DATA_DIR_INDEX_RESOURCE = 2
DATA_DIR_INDEX_EXCEPTION = 3
DATA_DIR_INDEX_CERTIFICATE = 4
DATA_DIR_INDEX_BASE_RELOC = 5
DATA_DIR_INDEX_DEBUG = 6
DATA_DIR_INDEX_ARCHITECTURE = 7
DATA_DIR_INDEX_GLOBAL_PTR = 8
DATA_DIR_INDEX_TLS = 9
DATA_DIR_INDEX_LOAD_CONFIG = 10
DATA_DIR_INDEX_BOUND_IMPORT = 11
DATA_DIR_INDEX_IMPORT_ADDRESS = 12
DATA_DIR_INDEX_DELAY_IMPORT_DESCRIPTOR = 13
DATA_DIR_INDEX_CLR_HEADER = 14
DEFAULT_DATA_DIRECTORIES = [dict(VirtualAddress=0, Size=0)] * 16

# IMAGE_OPTIONAL_HEADER.DllCharacteristics
IMAGE_DLLCHARACTERISTICS_DYNAMIC_BASE = 'IMAGE_DLLCHARACTERISTICS_DYNAMIC_BASE'
IMAGE_DLLCHARACTERISTICS_FORCE_INTEGRITY = 'IMAGE_DLLCHARACTERISTICS_FORCE_INTEGRITY'
IMAGE_DLLCHARACTERISTICS_NX_COMPAT = 'IMAGE_DLLCHARACTERISTICS_NX_COMPAT'
IMAGE_DLLCHARACTERISTICS_NO_ISOLATION = 'IMAGE_DLLCHARACTERISTICS_NO_ISOLATION'
IMAGE_DLLCHARACTERISTICS_NO_SEH = 'IMAGE_DLLCHARACTERISTICS_NO_SEH'
IMAGE_DLLCHARACTERISTICS_NO_BIND = 'IMAGE_DLLCHARACTERISTICS_NO_BIND'
IMAGE_DLLCHARACTERISTICS_WDM_DRIVER = 'IMAGE_DLLCHARACTERISTICS_WDM_DRIVER'
IMAGE_DLLCHARACTERISTICS_TERMINAL_SERVER_AWARE = 'IMAGE_DLLCHARACTERISTICS_TERMINAL_SERVER_AWARE'

# IMAGE_FILE_HEADER.Machine
IMAGE_FILE_MACHINE_UNKNOWN = 0x0
IMAGE_FILE_MACHINE_AM33 = 0x1d3
IMAGE_FILE_MACHINE_AMD64 = 0x8664
IMAGE_FILE_MACHINE_ARM = 0x1c0
IMAGE_FILE_MACHINE_ARM64 = 0xaa64
IMAGE_FILE_MACHINE_ARMNT = 0x1c4
IMAGE_FILE_MACHINE_EBC = 0xebc
IMAGE_FILE_MACHINE_I386 = 0x14c
IMAGE_FILE_MACHINE_IA64 = 0x200
IMAGE_FILE_MACHINE_M32R = 0x9041
IMAGE_FILE_MACHINE_MIPS16 = 0x266
IMAGE_FILE_MACHINE_MIPSFPU = 0x366
IMAGE_FILE_MACHINE_MIPSFPU16 = 0x466
IMAGE_FILE_MACHINE_POWERPC = 0x1f0
IMAGE_FILE_MACHINE_POWERPCFP = 0x1f1
IMAGE_FILE_MACHINE_R4000 = 0x166
IMAGE_FILE_MACHINE_RISCV32 = 0x5032
IMAGE_FILE_MACHINE_RISCV64 = 0x5064
IMAGE_FILE_MACHINE_RISCV128 = 0x5128
IMAGE_FILE_MACHINE_SH3 = 0x1a2
IMAGE_FILE_MACHINE_SH3DSP = 0x1a3
IMAGE_FILE_MACHINE_SH4 = 0x1a6
IMAGE_FILE_MACHINE_SH5 = 0x1a8
IMAGE_FILE_MACHINE_THUMB = 0x1c2
IMAGE_FILE_MACHINE_WCEMIPSV2 = 0x169


# IMAGE_FILE_HEADER characterstics.
IMAGE_FILE_RELOCS_STRIPPED = 'IMAGE_FILE_RELOCS_STRIPPED'
IMAGE_FILE_EXECUTABLE_IMAGE = 'IMAGE_FILE_EXECUTABLE_IMAGE'
IMAGE_FILE_LINE_NUMS_STRIPPED = 'IMAGE_FILE_LINE_NUMS_STRIPPED'
IMAGE_FILE_LOCAL_SYMS_STRIPPED = 'IMAGE_FILE_LOCAL_SYMS_STRIPPED'
IMAGE_FILE_AGGRESIVE_WS_TRIM = 'IMAGE_FILE_AGGRESIVE_WS_TRIM'
IMAGE_FILE_LARGE_ADDRESS_AWARE = 'IMAGE_FILE_LARGE_ADDRESS_AWARE'
IMAGE_FILE_BYTES_REVERSED_LO = 'IMAGE_FILE_BYTES_REVERSED_LO'
IMAGE_FILE_32BIT_MACHINE = 'IMAGE_FILE_32BIT_MACHINE'
IMAGE_FILE_DEBUG_STRIPPED = 'IMAGE_FILE_DEBUG_STRIPPED'
IMAGE_FILE_REMOVABLE_RUN_FROM_SWAP = 'IMAGE_FILE_REMOVABLE_RUN_FROM_SWAP'
IMAGE_FILE_NET_RUN_FROM_SWAP = 'IMAGE_FILE_NET_RUN_FROM_SWAP'
IMAGE_FILE_SYSTEM = 'IMAGE_FILE_SYSTEM'
IMAGE_FILE_DLL = 'IMAGE_FILE_DLL'
IMAGE_FILE_UP_SYSTEM_ONLY = 'IMAGE_FILE_UP_SYSTEM_ONLY'
IMAGE_FILE_BYTES_REVERSED_HI = 'IMAGE_FILE_BYTES_REVERSED_HI'


import abc
import logging
import six

# This is here for type hints and autocomplete in PyCharm
# noinspection PyUnreachableCode
if False:
    from mwcp import FileObject

logger = logging.getLogger(__name__)


# A way to create a class properties
# (Adding ABCMeta so, parsers have the freedom to use it.)
class ParserMeta(abc.ABCMeta):
    @property
    def name(cls):
        try:
            return cls._name
        except AttributeError:
            return cls.__name__

    @name.setter
    def name(cls, value):
        cls._name = value

    @property
    def source(cls):
        try:
            return cls._source
        except AttributeError:
            module, _, _ = cls.__module__.partition(".")
            return module

    @source.setter
    def source(cls, value):
        cls._source = value

    def __repr__(cls):
        return "<{}>".format(cls.name)


@six.add_metaclass(ParserMeta)
class Parser(object):
    """
    Interface for all parser objects.
    Either use this as a base for all component parsers, or
    inherit this class into a customized base class for all parsers.  This class includes some of the required data
    used by various other classes.
    """

    file_object = None  # type: FileObject
    # This is the description that will be given the the file object during output
    # if no description is set in the file_object. This must be overwritten by inherited classes.
    DESCRIPTION = None

    # TODO: Deprecate the AUTHOR field?
    AUTHOR = ""  # Optional

    def __init__(self, file_object, reporter, dispatcher):
        """
        Initializes the Parser.

        :param FileObject file_object: Object containing data about component file.
        :param mwcp.Reporter reporter: reference to reporter object that executed this parser.
        :param Dispatcher dispatcher: reference to the dispatcher object
        """
        if not self.DESCRIPTION:
            raise NotImplementedError("Parser class is missing a DESCRIPTION.")
        self.file_object = file_object
        self.reporter = reporter
        self.dispatcher = dispatcher
        self.logger = logging.getLogger(".".join([self.__class__.__module__, self.__class__.__name__]))

    @classmethod
    def get_logger(cls):
        return logging.getLogger(".".join([cls.__module__, cls.__name__]))

    @classmethod
    def iter_subclasses(cls):
        """Yields all classes that inherit from this class."""
        for subclass in cls.__subclasses__():
            yield subclass
            for _subclass in subclass.iter_subclasses():
                yield _subclass

    @classmethod
    def identify(cls, file_object):
        """
        Determines if this parser is identified to support the given file_object.
        This function must be overwritten in order to support identification.

        The passed in file_object may be modified at this time to provide
        a new file_name or description.
        (Be aware, that this change will be in affect for future parsers.
        Therefore, don't change it if you are returning False or the dispatcher is in greedy mode.)

        :param file_object: file object to use for identification
        :type file_object: dispatcher.FileObject

        :return bool: Boolean indicating if this parser supports the file_object
        """
        logger.warning("Missing identify() function for: {}.{}".format(cls.__module__, cls.__name__))
        return True  # Default to True to keep backwards compatibility for legacy parsers.

    @classmethod
    def parse(cls, file_object, reporter, dispatcher=None):
        """
        Runs parser on given file_object.

        :param FileObject file_object: Object containing data about component file.
        :param mwcp.Reporter reporter: reference to reporter object that executed this parser.
        :param Dispatcher dispatcher: reference to the dispatcher object. (if used)
        :return:
        """
        # TODO: Use reporter to output metadata about initial input file.
        if dispatcher:
            parser_object = cls(file_object, reporter, dispatcher)
            parser_object.run()

        # If dispatcher isn't provided, create a dummy one containing only this parser.
        else:
            from mwcp import Dispatcher  # Must import here to avoid cyclic import.

            dispatcher = Dispatcher(cls.name, author=cls.AUTHOR, description=cls.DESCRIPTION, parsers=[cls])
            dispatcher.parse(file_object, reporter)

    def run(self):
        """
        This function can be overwritten. It is called to run the parser.
        You don't have to overwrite this method if you only want to identify/output the file.
        :return:
        """
        pass



"""
pecon - PE file reCONstructor

Usage:
    >>> from mwcp.utils import pecon

    # Create a PE object
    >>> pe = pecon.PE()

    # Fill in pe with known information (fields not provided will contain a default as defined in the PE constructor)
    >>> pe.DosHeader.e_lfanew = 0x3211
    >>> pe.OptionalHeader.SizeOfCode = 0x3141241
    >>> pe.OptionalHeader.AddressOfEntryPoint = 0x43222
    >>> pe.OptionalHeader.Subsystem = pecon.IMAGE_SUBSYSTEM_WINDOWS_GUI
    # DataDirectory is a list of IMAGE_DATA_DIRECTORY structs.
    # By default it contains the standard 16, which can be accessed with indexes or though helper attributes
    # ("imports", "exports", etc)
    # NOTE: While it would make more sense to call it "DataDirectories", we are trying to be
    #       consistent with Microsoft's names.
    >>> pe.OptionalHeader.DataDirectory.imports.VirtualAddress = 0x101
    >>> pe.OptionalHeader.DataDirectory.imports.Size = 20
    # Create pecon.Section() objects to fill in section information.
    # (By default there are no sections.)
    >>> pe.SectionTable.append(pecon.Section(Name='.text', VirtualSize=4, VirtualAddress=0x3422, data=b'blah'))

    # Generate file data.
    >>> pe_data = pe.build()
    # To only build the header, you can tell it to avoid writing the section data.
    >>> pe_data = pe.build(write_section_data=False)

    # You can also modifiy fields in an already existing exe file.
    >>> pe = pecon.PE(exe_data)
    >>> pe.OptionalHeader.SizeOfCode = 0x3422
    >>> pe_data = pe.build()


"""

import copy
import io
import logging

logger = logging.getLogger(__name__)

from mwcp.utils import construct
from construct import this

# Expose the constants users will need.
from mwcp.utils.construct.windows_constants import *


# Overwrite Container class to provide deepcopy functionality.
class Container(construct.Container):

    @classmethod
    def from_container(cls, container_object):
        """Factory method for converting an already existing Container object."""
        _dict = {}
        for key, value in container_object.items():
            if isinstance(value, dict):
                value = cls.from_container(value)
            if isinstance(value, list):
                for i in range(len(value)):
                    # one level is all that is necessary for what we are doing.
                    if isinstance(value[i], dict):
                        value[i] = cls.from_container(value[i])
            _dict[key] = value
        return cls(_dict)

    def __deepcopy__(self, memo):
        _copy = Container()
        for key, value in self.items():
            _copy[key] = copy.deepcopy(value, memo)
        return _copy


class DataDirectories(construct.ListContainer):
    """
    A list of IMAGE_DATA_DIRECTORY entries

    Provides convenience properties for accessing standard directories by name.

    :param int size: Number of directory entries. defaults to the standard size of 16
    """

    def __init__(self, size=16):
        super(DataDirectories, self).__init__()
        for _ in range(size):
            self.append(Container(VirtualAddress=0, Size=0))

    def sizeof(self):
        return construct.IMAGE_DATA_DIRECTORY.sizeof() * len(self)

    # Provide convenience properties for accessing standard data directories.

    @property
    def exports(self):
        return self[construct.DATA_DIR_INDEX_EXPORTS]

    @exports.setter
    def exports(self, value):
        self[construct.DATA_DIR_INDEX_EXPORTS] = value

    @property
    def imports(self):
        return self[construct.DATA_DIR_INDEX_IMPORTS]

    @imports.setter
    def imports(self, value):
        self[construct.DATA_DIR_INDEX_IMPORTS] = value

    @property
    def resource(self):
        return self[construct.DATA_DIR_INDEX_RESOURCE]

    @resource.setter
    def resource(self, value):
        self[construct.DATA_DIR_INDEX_RESOURCE] = value

    @property
    def exception(self):
        return self[construct.DATA_DIR_INDEX_EXCEPTION]

    @exception.setter
    def exception(self, value):
        self[construct.DATA_DIR_INDEX_EXCEPTION] = value

    @property
    def certificate(self):
        return self[construct.DATA_DIR_INDEX_CERTIFICATE]

    @certificate.setter
    def certificate(self, value):
        self[construct.DATA_DIR_INDEX_CERTIFICATE] = value

    @property
    def base_reloc(self):
        return self[construct.DATA_DIR_INDEX_BASE_RELOC]

    @base_reloc.setter
    def base_reloc(self, value):
        self[construct.DATA_DIR_INDEX_BASE_RELOC] = value

    @property
    def debug(self):
        return self[construct.DATA_DIR_INDEX_DEBUG]

    @debug.setter
    def debug(self, value):
        self[construct.DATA_DIR_INDEX_DEBUG] = value

    @property
    def architecture(self):
        return self[construct.DATA_DIR_INDEX_ARCHITECTURE]

    @architecture.setter
    def architecture(self, value):
        self[construct.DATA_DIR_INDEX_ARCHITECTURE] = value

    @property
    def global_ptr(self):
        return self[construct.DATA_DIR_INDEX_GLOBAL_PTR]

    @global_ptr.setter
    def global_ptr(self, value):
        self[construct.DATA_DIR_INDEX_GLOBAL_PTR] = value

    @property
    def tls(self):
        return self[construct.DATA_DIR_INDEX_TLS]

    @tls.setter
    def tls(self, value):
        self[construct.DATA_DIR_INDEX_TLS] = value

    @property
    def load_config(self):
        return self[construct.DATA_DIR_INDEX_LOAD_CONFIG]

    @load_config.setter
    def load_config(self, value):
        self[construct.DATA_DIR_INDEX_LOAD_CONFIG] = value

    @property
    def bound_import(self):
        return self[construct.DATA_DIR_INDEX_BOUND_IMPORT]

    @bound_import.setter
    def bound_import(self, value):
        self[construct.DATA_DIR_INDEX_BOUND_IMPORT] = value

    @property
    def import_address(self):
        return self[construct.DATA_DIR_INDEX_IMPORT_ADDRESS]

    @import_address.setter
    def import_address(self, value):
        self[construct.DATA_DIR_INDEX_IMPORT_ADDRESS] = value

    @property
    def dely_import_descriptor(self):
        return self[construct.DATA_DIR_INDEX_DELAY_IMPORT_DESCRIPTOR]

    @dely_import_descriptor.setter
    def dely_import_descriptor(self, value):
        self[construct.DATA_DIR_INDEX_DELAY_IMPORT_DESCRIPTOR] = value

    @property
    def clr_header(self):
        return self[construct.DATA_DIR_INDEX_CLR_HEADER]

    @clr_header.setter
    def clr_header(self, value):
        self[construct.DATA_DIR_INDEX_CLR_HEADER] = value


class Section(Container):
    """
    Container for IMAGE_SECTION_HEADER

    (Provides defaults for non-filled values.)
    """

    def __init__(self, *args, **kw):
        _section_header = {
            'Name': '',
            'VirtualSize': 0,
            'VirtualAddress': 0,
            'SizeOfRawData': 0,
            'PointerToRawData': 0,
            'PointerToRelocations': 0,
            'PointerToLinenumbers': 0,
            'NumberOfrelocations': 0,
            'NumberOfLinenumbers': 0,
            'Characteristics': [],
            'data': '',
        }
        super(Section, self).__init__(_section_header)
        for arg in args:
            self.update(arg)
        self.update(kw)


class PE(Container):

    def __init__(self, data=None, is_64bit=False):
        """

        :param data: Data from an existing PE file, if provided, this will be used as the base line.
        :param is_64bit: Whether to make a 64 bit or 32 bit PE file. Defaults to 32 bit.
            (NOTE: This is only applicable if not passing in data.)
        """

        super(PE, self).__init__()

        if data:
            # If user provided data, parse it and use it as a base point.
            self._parse(data)
            return

        # Otherwise create a default pe.

        _characteristics = [
            construct.IMAGE_FILE_RELOCS_STRIPPED,
            construct.IMAGE_FILE_EXECUTABLE_IMAGE,
            construct.IMAGE_FILE_LINE_NUMS_STRIPPED,
            construct.IMAGE_FILE_LOCAL_SYMS_STRIPPED,
        ]

        if is_64bit:
            _magic = construct.IMAGE_NT_OPTIONAL_HDR64_MAGIC
            _machine = construct.IMAGE_FILE_MACHINE_AMD64
        else:
            _magic = construct.IMAGE_NT_OPTIONAL_HDR32_MAGIC
            _machine = construct.IMAGE_FILE_MACHINE_I386
            _characteristics.append(construct.IMAGE_FILE_32BIT_MACHINE)

        _dos_header = Container({
            'e_magic': b'MZ',
            'e_cblp': 0x90,
            'e_cp': 0x03,
            'e_crlc': 0,
            'e_cparhdr': 4,
            'e_mimalloc': 0,
            'e_maxalloc': 0xffff,
            'e_ss': 0,
            'e_sp': 184,
            'e_csum': 0,
            'e_ip': 0,
            'e_cs': 0,
            'e_lfarlc': 64,
            'e_ovno': 0,
            'e_res1': b'\x00\x00\x00\x00\x00\x00\x00\x00',
            'e_oemid': 0,
            'e_oeminfo': 0,
            'e_res2': b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00',
            'e_lfanew': 224,
        })

        _data_directories = DataDirectories()

        _optional_header = Container({
            'Magic': _magic,
            'MajorLinkerVersion': 1,
            'MinorLinkerVersion': 71,
            'SizeOfCode': 0,
            'SizeOfInitializedData': 0,
            'SizeOfUninitializedData': 0,
            'AddressOfEntryPoint': 0,
            'BaseOfCode': 0,
            'BaseOfData': 0,
            'ImageBase': 0,
            'SectionAlignment': 4096,
            'FileAlignment': 512,
            'MajorOperatingSystemVersion': 1,
            'MinorOperatingSystemVersion': 0,
            'MajorImageVersion': 0,
            'MinorImageVersion': 0,
            'MajorSubsystemVersion': 5,
            'MinorSubsystemVersion': 1,
            'Win32VersionValue': 0,  # must be 0 (but I guess still allow them to change it)
            'SizeOfImage': 0,
            'SizeOfHeaders': 0,
            'CheckSum': 0,
            'Subsystem': construct.IMAGE_SUBSYSTEM_WINDOWS_CUI,
            'DllCharacteristics': [],
            'SizeOfStackReserve': 1048576,
            'SizeOfStackCommit': 4096,
            'SizeOfHeapReserve': 1048576,
            'SizeOfHeapCommit': 4096,
            'LoaderFlags': 0,
            'NumberOfRvaAndSizes': _data_directories.sizeof(),
            'DataDirectory': _data_directories,
        })

        _file_header = Container({
            'Machine': _machine,
            'NumberOfSections': 0,
            'TimeDateStamp': 0,
            'PointerToSymbolTable': 0,
            'NumberOfSymbols': 0,
            'SizeOfOptionalHeader': construct.IMAGE_OPTIONAL_HEADER.sizeof(**_optional_header),
            'Characteristics': _characteristics,
        })

        self.DosHeader = Container(_dos_header)

        # Default to "ret" opcode.
        self.DosStub = b'\xc3'

        self.NTHeaders = Container(
            Signature=0x4550,   # b'PE\x00\x00'
            FileHeader=Container(_file_header),
            OptionalHeader=Container(_optional_header)
        )

        self.SectionTable = construct.ListContainer()

    def _parse(self, data):
        """
        Parses data containing PE file and updates dictionary to reflect results.

        :raises ConstructError: If provided data could not be parsed (ie. not a pe)
        """
        pe = construct.PEFILE_HEADER.parse(data)

        # Convert Container classes to use ours, so we can deepcopy.
        pe = Container.from_container(pe)

        # Convert the FlagEnums into list of constants.
        for section in pe.SectionTable:
            if isinstance(section.Characteristics, dict):
                section.Characteristics = [flag for flag, value in section.Characteristics.items() if value]
            section.data = data[section.PointerToRawData:section.PointerToRawData + section.SizeOfRawData]
        file_header = pe.NTHeaders.FileHeader
        if isinstance(file_header.Characteristics, dict):
            file_header.Characteristics = [
                flag for flag, value in file_header.Characteristics.items() if value]
        optional_header = pe.NTHeaders.OptionalHeader
        if isinstance(optional_header.DllCharacteristics, dict):
            optional_header.DllCharacteristics = [
                flag for flag, value in optional_header.DllCharacteristics.items() if value]

        self.update(pe)

    def _fix_section(self, section):
        """Fixes up section container for building."""
        section_copy = copy.deepcopy(section)

        # Formulate characteristics based on name if they weren't provided.
        if not section_copy.Characteristics:
            _characteristics = [construct.IMAGE_SCN_MEM_READ]
            if section_copy.Name == u'.text':
                _characteristics += [construct.IMAGE_SCN_CNT_CODE, construct.IMAGE_SCN_MEM_EXECUTE]
            else:
                _characteristics += [construct.IMAGE_SCN_CNT_INITIALIZED_DATA]
            if section_copy.Name == u'.data':
                _characteristics += [construct.IMAGE_SCN_MEM_WRITE]
            if section_copy.Name == u'.reloc':
                _characteristics += [construct.IMAGE_SCN_MEM_DISCARDABLE]
            section_copy.Characteristics = _characteristics

        # FlagEnums must be a dictionary.
        if isinstance(section_copy.Characteristics, list):
            section_copy.Characteristics = {flag: True for flag in section_copy.Characteristics}

        # Fix up data to be consistent.
        data_size = max(section_copy.SizeOfRawData, len(section_copy.data))
        section_copy.SizeOfRawData = data_size
        section_copy.data = section_copy.data.ljust(data_size, b'\x00')[:data_size]

        return section_copy

    def build(self, write_section_data=True):
        """
        Generate PE file.

        :param write_section_data: Whether to include section data (otherwise only the headers are written)

        :returns bytes: PE file data.

        :raises ValueError: If set attributes contains contradicting data.
        """
        pe = copy.deepcopy(self)

        # Pad dos stub to match e_lfanew. (Warn if dos stub is too large.)
        dos_stub_size = pe.DosHeader.e_lfanew - construct.IMAGE_DOS_HEADER.sizeof()
        if len(pe.DosStub) > dos_stub_size:
            raise ValueError(
                'Provided DOS stub is too large for provided DosHeader.e_lfanew: {}'.format(pe.DosHeader.e_lfanew))
        pe.DosStub = pe.DosStub.ljust(dos_stub_size, b'\x00')

        # Fix file header.
        file_header = pe.NTHeaders.FileHeader
        if file_header.NumberOfSections and file_header.NumberOfSections != len(pe.SectionTable):
            logger.debug(
                'NTHeaders.FileHeader.NumberOfSections does not equal the number of sections provided. Auto-adjusting.')
        file_header.NumberOfSections = len(pe.SectionTable)
        if isinstance(file_header.Characteristics, list):
            file_header.Characteristics = {flag: True for flag in file_header.Characteristics}

        # Fix sections.
        pe.SectionTable = list(map(self._fix_section, pe.SectionTable))

        # Fix optional header.
        optional_header = pe.NTHeaders.OptionalHeader
        number_of_rva_and_sizes = len(optional_header.DataDirectory)
        optional_header.NumberOfRvaAndSizes = number_of_rva_and_sizes
        file_header.SizeOfOptionalHeader = construct.IMAGE_OPTIONAL_HEADER.sizeof(**optional_header)
        if isinstance(optional_header.DllCharacteristics, list):
            optional_header.DllCharacteristics = {flag: True for flag in optional_header.DllCharacteristics}

        # SizeOfHeaders is the sum of the headers rounded by FileAlignment.
        headers_size = construct.PEFILE_HEADER.sizeof(**pe)
        file_alignment = optional_header.FileAlignment - 1
        headers_size = (headers_size + file_alignment) & 0xffffffff - file_alignment
        optional_header.SizeOfHeaders = headers_size

        pe_data = construct.PEFILE_HEADER.build(pe)

        # Add section data.
        if write_section_data:
            stream = io.BytesIO(pe_data)
            spec = construct.Pointer(this.PointerToRawData, construct.Bytes(this.SizeOfRawData))
            for section in pe.SectionTable:
                spec.build_stream(section.data, stream, **section)
            pe_data = stream.getvalue()

        return pe_data

    # Provide convenience properties for accessing NTHeader components.

    @property
    def OptionalHeader(self):
        return self.NTHeaders.OptionalHeader

    @OptionalHeader.setter
    def OptionalHeader(self, value):
        self.NTHeaders.OptionalHeader = value

    @property
    def FileHeader(self):
        return self.NTHeaders.FileHeader

    @FileHeader.setter
    def FileHeader(self, value):
        self.NTHeaders.FileHeader = value

    @property
    def Signature(self):
        return self.NTHeaders.Signature

    @Signature.setter
    def Signature(self, value):
        self.NTHeaders.Signature = value



"""
Implements FileObject class used to provide an interface for the file being parsed.
"""

import hashlib
import io
import logging
import os
from typing import List

import pefile
from mwcp.utils import elffileutils, pefileutils

try:
    import kordesii
except ImportError:
    # Kordesii support is optional.
    kordesii = None

from mwcp.utils.stringutils import convert_to_unicode

logger = logging.getLogger(__name__)


class FileObject(object):
    """
    This class represents a file object which is to be parsed by the MWCP parser.
    It is pushed into the dispatcher queue for processing.
    """

    def __init__(
        self,
        file_data: bytes,
        reporter,
        pe: pefile.PE = None,
        file_name=None,
        def_stub=None,
        description=None,
        output_file=True,
        use_supplied_fname=True,
        use_arch=False,
        ext=".bin",
    ):
        """
        Initializes the FileObject.

        :param bytes file_data: Data for the file.
        :param pefile.PE pe: PE object for the file.
        :param mwcp.Reporter reporter: MWCP reporter.
        :param str file_name: File name to use if file is not a PE or use_supplied_fname was specified.
        :param str description: Description of the file object.
        :param bool output_file: Boolean indicating if file should be outputted when the dispatcher process the file.
        :param bool use_supplied_fname: Boolean indicating if the file_name should be used even if the file is a PE.
        :param str def_stub: def_stub argument to pass to obtain_original_filename()
        :param bool use_arch: use_arch argument to pass to obtain_original_filename()
        :param str ext: default extension to use if not determined from pe file.
        """
        # Ensure we are getting a bytes string. Libraries like pefile depend on this.
        if not isinstance(file_data, bytes):
            raise TypeError("file_data must be a bytes string.")

        self._file_path = None
        self._md5 = None
        self._sha1 = None
        self._sha256 = None
        self._stack_strings = None
        self._static_strings = None
        self._resources = None
        self._elf = None
        self._elf_attempt = False
        self.output_file = output_file
        self._outputted_file = False
        self._kordesii_cache = {}
        self.parent = None  # Parent FileObject from which FileObject was extracted from (this is set externally).
        self.parser = None  # This will be set by the dispatcher.
        self.children = []  # List of residual FileObject
        self.file_data = file_data
        self.reporter = reporter
        self.description = description
        self.knowledge_base = {}

        self.pe = pe or pefileutils.obtain_pe(file_data)

        use_supplied_fname = use_supplied_fname or not self.pe

        if file_name and use_supplied_fname:
            self._file_name = file_name
        else:
            self._file_name = pefileutils.obtain_original_filename(
                def_stub or self.md5, pe=self.pe, use_arch=use_arch, ext=ext
            )
        self._file_name = convert_to_unicode(self._file_name)

    def __enter__(self):
        """
        This allows us to use the file_data as a file-like object when used as a context manager.

        e.g.
            >> file_object = FileObject('hello world', None)
            >> with file_object as fo:
            ..     _ = fo.seek(6)
            ..     print fo.read()
            world
        """
        self._open_file = io.BytesIO(self.file_data)
        return self._open_file

    def __exit__(self, *args):
        self._open_file.close()

    @property
    def siblings(self) -> List["FileObject"]:
        """List of FileObjects that came from the same parent."""
        if not self.parent:
            return []
        return [fo for fo in self.parent.children if fo is not self]

    # TODO: Deprecate "file_data" name in exchange for "data"?
    @property
    def data(self):
        """Just an alias for file_data"""
        return self.file_data

    @property
    def elf(self):
        """Returns elftools.ELFFile object or None if not an ELF file."""
        if not self._elf and not self._elf_attempt:
            self._elf_attempt = True
            self._elf = elffileutils.obtain_elf(self.file_data)
        return self._elf

    # TODO: Deprecate "file_name" name in exhange for "name"?
    @property
    def file_name(self):
        return self._file_name

    @file_name.setter
    def file_name(self, value):
        # If someone changes the name, record the rename.
        value = convert_to_unicode(value)
        if self._file_name != value:
            logger.info("Renamed {} to {}".format(self._file_name, value))
        self._file_name = value

    @property
    def name(self):
        return self.file_name

    @name.setter
    def name(self, value):
        self.file_name = value

    @property
    def parser_history(self):
        """
        Returns a history of the parser classes (including current) that has lead to the creation of the file object.
        e.g. [MalwareDropper, MalwareLoader, MalwareImplant]
        :return list: List of parser classes.
        """
        history = [self.parser]
        parent = self.parent
        while parent:
            history.append(parent.parser)
            parent = parent.parent
        return reversed(history)

    @property
    def md5(self):
        """
        Returns md5 hash of file.
        :return: hash of the file as a hex string
        """
        if not self._md5:
            self._md5 = hashlib.md5(self.file_data).hexdigest()
        return self._md5

    @property
    def sha1(self):
        """
        Returns sha1 hash of file.
        :return: hash of the file as a hex string
        """
        if not self._sha1:
            self._sha1 = hashlib.sha1(self.file_data).hexdigest()
        return self._sha1

    @property
    def sha256(self):
        """
        Returns sha256 hash of file.
        :return: hash of the file as a hex string
        """
        if not self._sha256:
            self._sha256 = hashlib.sha256(self.file_data).hexdigest()
        return self._sha256

    @property
    def file_path(self):
        """
        Returns a full file path to the file object.
        This is useful for when you want to use this file on libraries which require
        a file path instead of data or file-like object (e.g. cabinet).
        Always create a temporary file, this avoids issues where the identify function requires the file_path and
        the file would be output before a description is set.
        """
        if not self._file_path:
            safe_file_name = convert_to_unicode(self.md5)
            file_path = os.path.join(self.reporter.managed_tempdir, safe_file_name)
            with open(file_path, "wb") as file_object:
                file_object.write(self.file_data)
            self._file_path = file_path

        return self._file_path

    @file_path.setter
    def file_path(self, value):
        """
        Setter for the file_path attribute. This is used if an external entity can
        provided a valid file_path.
        """
        self._file_path = value

    @property
    def stack_strings(self):
        """
        Returns the stack strings for the file.
        """
        if not self._stack_strings:
            kordesii_reporter = self.run_kordesii_decoder("stack_string")
            self._stack_strings = kordesii_reporter.get_strings()
        return self._stack_strings

    # TODO: Create a static_strings property?

    @property
    def resources(self):
        """Returns a list of the PE resources for the given file."""
        if self.pe and not self._resources:
            self._resources = list(pefileutils.iter_rsrc(self.pe))
        return self._resources

    @property
    def is_64bit(self):
        """
        Evaluates whether the file is a 64 bit pe file.

        :return: True if 64-bit, False if 32-bit, None if could not be determined.
        """
        if not self.pe:
            return None
        return pefileutils.is_64bit(pe=self.pe)

    def output(self):
        """
        Outputs FileObject instance to reporter it it hasn't already been outputted.
        """
        # Output file if we are allowed to and the file hasn't already been outputted.
        if self.output_file and not self._outputted_file:
            self.reporter.output_file(
                data=self.file_data, filename=self.file_name or "", description=self.description or ""
            )
            self._outputted_file = True

    def run_kordesii_decoder(self, decoder_name: str, warn_no_strings=True, **run_config):
        """
        Run the specified kordesii decoder against the file data.  The reporter object is returned
        and can be accessed as necessary to obtain output files, etc.

        :param decoder_name: name of the decoder to run
        :param warn_no_strings: Whether to produce a warning if no string were found.
        :param run_config: Run configuration options to pass along to kordesii.run_ida()

        :return: Instance of the kordesii_reporter.

        :raises RuntimeError: If kordesii is not installed.
        """
        if not kordesii:
            raise RuntimeError("Please install kordesii to use this function.")

        # Pull from cache if we already ran this decoder.
        if decoder_name in self._kordesii_cache:
            return self._kordesii_cache[decoder_name]

        logger.info(f"Running {decoder_name} kordesii decoder on file {self.file_name}.")
        # Ensure decoderdir sources are populated
        kordesii.register_entry_points()

        kordesii_reporter = kordesii.Reporter(base64outputfiles=True)

        if "log" not in run_config:
            run_config["log"] = True
        kordesii_reporter.run_decoder(decoder_name, data=self.file_data, **run_config)

        if warn_no_strings:
            decrypted_strings = kordesii_reporter.get_strings()
            if not decrypted_strings:
                # Not necessarily a bad thing, the decoder might be used for something else.
                logger.info(f"No decrypted strings were returned by the decoder for file {self.file_name}.")

        # Cache results
        self._kordesii_cache[decoder_name] = kordesii_reporter

        return kordesii_reporter



"""This is a wrapper interface to the construct library which adds extra helper functions."""

# from __future__ import absolute_import

# Import interface
from .version28 import *

from .construct_html import html_hex
from .helpers import *
from .dotnet import *
from .datetime_ import *
from .windows_structures import *
from .network import *
from .windows_enums import *
from . import ARM
from . import MIPS




"""
Date/Time constructs
"""

from __future__ import absolute_import

import datetime

from .version28 import *


# TODO: Implement _encode.
class _DateTimeDateDataAdapter(Adapter):
    r"""
    Adapter for a C# DateTime.dateData object to DateTime format. Obtain the DateTime.Ticks and the DateTime.Kind
    property to format datetime.


    >>> _DateTimeDateDataAdapter(Int64sl).parse('\x80\xb4N3\xd1\xd4\xd1H')
    '2014-11-23 01:09:01 UTC'
    """
    def _decode(self, obj, context, path):
        ticks = obj & 0x3fffffffffffffff
        kind = (obj >> 62) & 0x03
        converted_ticks = datetime.datetime(1, 1, 1) + datetime.timedelta(microseconds=ticks / 10)
        if kind == 0:
            return converted_ticks.strftime("%Y-%m-%d %H:%M:%S")
        elif kind == 1:
            return converted_ticks.strftime("%Y-%m-%d %H:%M:%S UTC")
        elif kind == 2:
            return converted_ticks.strftime("%Y-%m-%d %H:%M:%S Local")


DateTimeDateData = _DateTimeDateDataAdapter(Int64sl)


# TODO: Implement _encode
class _EpochTimeAdapter(Adapter):
    r"""
    Adapter to convert time_t, EpochTime, to an isoformat

    >>> _EpochTimeAdapter(Int32ul).parse('\xff\x93\x37\x57')
    '2016-05-14T17:09:19'
    """
    def _decode(self, obj, context, path):
        return datetime.datetime.fromtimestamp(obj).isoformat()

# Hide the adapter
EpochTime = _EpochTimeAdapter(Int32ul)